import functools
import inspect
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, TypeVar

import jax
import numpy as np

from ciclo.logging import History, Logs
from ciclo.loops.loop import LoopFunctionCallback
from ciclo.timetracking import Period
from ciclo.types import A, B, Batch

F = TypeVar("F", bound=Callable[..., Any])


def logs(*args, **kwargs: Dict[str, Any]) -> Logs:
    return Logs(*args, **kwargs)


def history(logs_list: Optional[List[Logs]] = None) -> History:
    if logs_list is None:
        return History()

    return History(logs_list)


def at(
    steps: Optional[int] = None,
    *,
    samples: Optional[int] = None,
    time: Optional[float] = None,
    date: Optional[float] = None,
) -> Period:
    return Period.create(steps=steps, samples=samples, time=time, date=date)


def is_scalar(x):
    if isinstance(x, (int, float, bool)):
        return True
    elif hasattr(x, "shape"):
        return x.shape == () or x.shape == (1,)
    else:
        return False


def callback(f) -> LoopFunctionCallback:
    return LoopFunctionCallback(f)


def get_batch_size(
    batch: Batch, batch_size_fn: Callable[[List[Tuple[int, ...]]], int]
) -> int:
    def get_shape(x) -> Tuple[int, ...]:
        if not hasattr(x, "shape"):
            return (1,)
        shape = x.shape
        if len(shape) == 0:
            shape = (1,)
        return shape

    shapes = [get_shape(x) for x in jax.tree_util.tree_leaves(batch)]

    if len(shapes) == 0:
        return 1

    return batch_size_fn(shapes)


def max_first_axis(shapes: List[Tuple[int, ...]]) -> int:
    return max(s[0] for s in shapes)


def inject(f: Callable[..., A]) -> Callable[..., A]:
    @functools.wraps(f)
    def _inject(*args) -> A:
        n_args = len(inspect.getfullargspec(f).args)
        if inspect.ismethod(f) or inspect.ismethod(f.__call__):
            n_args -= 1
        return f(*args[:n_args])

    return _inject



from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from ciclo import callbacks as callbacks_lib
from ciclo import schedules
from ciclo.logging import History
from ciclo.loops import loop
from ciclo.loops.loop import LoopCallbackLike, LoopOutput
from ciclo.schedules import ScheduleLike
from ciclo.timetracking import Elapsed, Period, PeriodLike
from ciclo.types import B, S

CallbackOrList = Union[LoopCallbackLike, List[LoopCallbackLike]]
FitScheduleLike = Union[ScheduleLike, str]
FitInputTasks = Dict[FitScheduleLike, CallbackOrList]
FitCallback = Any

PREDICT_NAMES = {
    (ON_PREDICT_STEP := "on_predict_step"),
    # also support the simple names
    (PREDICT_STEP := "predict_step"),
    # ----------------
    (ON_PREDICT_BATCH_BEGIN := "on_predict_batch_begin"),
    (ON_PREDICT_BATCH_END := "on_predict_batch_end"),
    (ON_PREDICT_BEGIN := "on_predict_begin"),
    (ON_PREDICT_END := "on_predict_end"),
}

TEST_NAMES = {
    (ON_TEST_STEP := "on_test_step"),
    (ON_RESET_STEP := "on_reset_step"),
    # also support the simple names
    (TEST_STEP := "test_step"),
    (RESET_STEP := "reset_step"),
    # ----------------
    (ON_TEST_BATCH_BEGIN := "on_test_batch_begin"),
    (ON_TEST_BATCH_END := "on_test_batch_end"),
    (ON_TEST_BEGIN := "on_test_begin"),
    (ON_TEST_END := "on_test_end"),
}

# NOTE: due to a choice in the implementation of the fit_loop
# there is no ON_EPOCH_BEGIN, instead you can use a combination
# of ON_TRAIN_BEGIN and ON_EPOCH_END
FIT_NAMES = {
    (ON_TRAIN_STEP := "on_train_step"),
    ON_RESET_STEP,
    # also support the simple names
    (TRAIN_STEP := "train_step"),
    RESET_STEP,
    # ----------------
    (ON_TRAIN_BEGIN := "on_train_begin"),
    (ON_TRAIN_END := "on_train_end"),
    # ON_EPOCH_BEGIN := "on_epoch_begin",
    (ON_EPOCH_END := "on_epoch_end"),
    (ON_TRAIN_BATCH_BEGIN := "on_train_batch_begin"),
    (ON_TRAIN_BATCH_END := "on_train_batch_end"),
}


def train_loop(
    state: S,
    dataset: Iterable[B],
    tasks: Optional[FitInputTasks] = None,
    *,
    callbacks: Optional[List[FitCallback]] = None,
    test_dataset: Optional[Callable[[], Iterable[B]]] = None,
    epoch_duration: Optional[PeriodLike] = None,
    test_duration: Optional[PeriodLike] = None,
    test_name: str = "test",
    stop: Optional[PeriodLike] = None,
    history: Optional[History] = None,
    elapsed: Optional[Elapsed] = None,
    catch_keyboard_interrupt: bool = True,
    metadata: Optional[Any] = None,
    batch_size_fn: Optional[Callable[[List[Tuple[int, ...]]], int]] = None,
) -> LoopOutput[S]:
    if tasks is None:
        tasks = {}

    if isinstance(epoch_duration, int):
        epoch_duration = Period.create(steps=epoch_duration)

    if isinstance(test_duration, int):
        test_duration = Period.create(steps=test_duration)

    additionl_tasks: Dict[ScheduleLike, CallbackOrList] = {}
    named_tasks: Dict[str, CallbackOrList] = {}
    for schedule in list(tasks.keys()):
        if isinstance(schedule, str):
            named_tasks[schedule] = tasks.pop(schedule)
        else:
            additionl_tasks[schedule] = tasks.pop(schedule)

    # check that all named tasks are valid
    unknown_names = set(named_tasks.keys()) - FIT_NAMES - TEST_NAMES
    if len(unknown_names) > 0:
        raise ValueError(f"Unknown named tasks: {unknown_names}")

    named_tasks = {
        schedule: callbacks if isinstance(callbacks, list) else [callbacks]
        for schedule, callbacks in named_tasks.items()
    }

    # extract callbacks from state
    for name in FIT_NAMES | TEST_NAMES:
        if hasattr(state, name):
            # note: here we get an unbounded method
            callback = getattr(type(state), name)
            named_tasks.setdefault(name, []).insert(0, callback)

    # extract callbacks from callbacks
    if callbacks is not None:
        for callback in callbacks:
            for name in FIT_NAMES | TEST_NAMES:
                if hasattr(callback, name):
                    callback_method = getattr(callback, name)
                    named_tasks.setdefault(name, []).append(callback_method)

    # extract test named tasks
    named_tasks_test = {}
    for name in TEST_NAMES - FIT_NAMES:
        if name in named_tasks:
            named_tasks_test[name] = named_tasks.pop(name)

    train_tasks = {}

    train_tasks[schedules.every(1)] = [
        *named_tasks.get(ON_TRAIN_BATCH_BEGIN, []),
        *named_tasks.get(TRAIN_STEP, []),
        *named_tasks.get(ON_TRAIN_STEP, []),
    ]

    if epoch_duration is not None:
        test_tasks = []
        test_tasks += named_tasks.pop(RESET_STEP, [])
        test_tasks += named_tasks.pop(ON_RESET_STEP, [])
        if test_dataset is not None:
            test_tasks.append(
                callbacks_lib.inner_loop(
                    test_name,
                    lambda state: test_loop(
                        state,
                        test_dataset(),
                        tasks=named_tasks_test,
                        stop=test_duration,
                        batch_size_fn=batch_size_fn,
                    ),
                )
            )
        test_tasks += named_tasks.pop(ON_EPOCH_END, [])

        eval_schedule = schedules.every(
            steps=epoch_duration.steps,
            samples=epoch_duration.samples,
            time=epoch_duration.time,
        )

        train_tasks[eval_schedule] = test_tasks

    train_tasks[schedules.every(1)] = [
        *named_tasks.get(ON_TRAIN_BATCH_END, []),
    ]

    train_tasks.update(additionl_tasks)

    return loop.loop(
        state,
        dataset,
        tasks=train_tasks,
        on_start=named_tasks.get(ON_TRAIN_BEGIN, None),
        on_end=named_tasks.get(ON_TRAIN_END, None),
        stop=stop,
        history=history,
        elapsed=elapsed,
        catch_keyboard_interrupt=catch_keyboard_interrupt,
        metadata=metadata,
        batch_size_fn=batch_size_fn,
    )


def test_loop(
    state: S,
    dataset: Iterable[B],
    tasks: Optional[FitInputTasks] = None,
    *,
    callbacks: Optional[FitCallback] = None,
    stop: Optional[PeriodLike] = None,
    history: Optional[History] = None,
    elapsed: Optional[Elapsed] = None,
    catch_keyboard_interrupt: bool = True,
    metadata: Optional[Any] = None,
    batch_size_fn: Optional[Callable[[List[Tuple[int, ...]]], int]] = None,
) -> LoopOutput[S]:
    if tasks is None:
        tasks = {}

    # create a copy of tasks to avoid modifying the original
    tasks = tasks.copy()
    additionl_tasks: Dict[ScheduleLike, CallbackOrList] = {}
    named_tasks: Dict[str, CallbackOrList] = {}
    for schedule in list(tasks.keys()):
        if isinstance(schedule, str):
            named_tasks[schedule] = tasks.pop(schedule)
        else:
            additionl_tasks[schedule] = tasks.pop(schedule)

    # check that all named tasks are valid
    unknown_names = set(named_tasks.keys()) - TEST_NAMES
    if len(unknown_names) > 0:
        raise ValueError(f"Unknown named tasks: {unknown_names}")

    named_tasks = {
        schedule: callbacks if isinstance(callbacks, list) else [callbacks]
        for schedule, callbacks in named_tasks.items()
    }

    # extract callbacks from state
    for name in TEST_NAMES:
        if hasattr(state, name):
            # note: here we get an unbounded method
            callback = getattr(type(state), name)
            named_tasks.setdefault(name, []).insert(0, callback)

    # extract callbacks from callbacks
    if callbacks is not None:
        for callback in callbacks:
            for name in TEST_NAMES:
                if hasattr(callback, name):
                    callback_method = getattr(callback, name)
                    named_tasks.setdefault(name, []).append(callback_method)

    test_tasks = {}
    test_tasks[schedules.every(1)] = [
        *named_tasks.get(ON_TEST_BATCH_BEGIN, []),
        *named_tasks.get(TEST_STEP, []),
        *named_tasks.get(ON_TEST_STEP, []),
        *named_tasks.get(ON_TEST_BATCH_END, []),
    ]
    test_tasks.update(additionl_tasks)

    return loop.loop(
        state,
        dataset,
        tasks=test_tasks,
        on_start=named_tasks.get(ON_TEST_BEGIN, [])
        + named_tasks.get(RESET_STEP, [])
        + named_tasks.get(ON_RESET_STEP, []),
        on_end=named_tasks.get(ON_TEST_END, None),
        stop=stop,
        history=history,
        elapsed=elapsed,
        catch_keyboard_interrupt=catch_keyboard_interrupt,
        metadata=metadata,
        batch_size_fn=batch_size_fn,
    )


def predict_loop(
    state: S,
    dataset: Iterable[B],
    tasks: Optional[FitInputTasks] = None,
    *,
    callbacks: Optional[FitCallback] = None,
    stop: Optional[PeriodLike] = None,
    history: Optional[History] = None,
    elapsed: Optional[Elapsed] = None,
    catch_keyboard_interrupt: bool = True,
    metadata: Optional[Any] = None,
    batch_size_fn: Optional[Callable[[List[Tuple[int, ...]]], int]] = None,
) -> LoopOutput[S]:
    if tasks is None:
        tasks = {}

    additionl_tasks: Dict[ScheduleLike, CallbackOrList] = {}
    named_tasks: Dict[str, CallbackOrList] = {}
    for schedule in list(tasks.keys()):
        if isinstance(schedule, str):
            named_tasks[schedule] = tasks.pop(schedule)
        else:
            additionl_tasks[schedule] = tasks.pop(schedule)

    # check that all named tasks are valid
    unknown_names = set(named_tasks.keys()) - PREDICT_NAMES
    if len(unknown_names) > 0:
        raise ValueError(f"Unknown named tasks: {unknown_names}")

    named_tasks = {
        schedule: callbacks if isinstance(callbacks, list) else [callbacks]
        for schedule, callbacks in named_tasks.items()
    }

    # extract callbacks from state
    for name in PREDICT_NAMES:
        if hasattr(state, name):
            # note: here we get an unbounded method
            callback = getattr(type(state), name)
            named_tasks.setdefault(name, []).insert(0, callback)

    # extract callbacks from callbacks
    if callbacks is not None:
        for callback in callbacks:
            for name in PREDICT_NAMES:
                if hasattr(callback, name):
                    callback_method = getattr(callback, name)
                    named_tasks.setdefault(name, []).append(callback_method)

    predict_tasks = {}
    predict_tasks[schedules.every(1)] = [
        *named_tasks.get(ON_PREDICT_BATCH_BEGIN, []),
        *named_tasks.get(PREDICT_STEP, []),
        *named_tasks.get(ON_PREDICT_STEP, []),
        *named_tasks.get(ON_PREDICT_BATCH_END, []),
    ]
    predict_tasks.update(additionl_tasks)

    return loop.loop(
        state,
        dataset,
        tasks=predict_tasks,
        on_start=named_tasks.get(ON_PREDICT_BEGIN, None),
        on_end=named_tasks.get(ON_PREDICT_END, None),
        stop=stop,
        history=history,
        elapsed=elapsed,
        catch_keyboard_interrupt=catch_keyboard_interrupt,
        metadata=metadata,
        batch_size_fn=batch_size_fn,
    )



from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Mapping,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

import jax
import optax
from flax import struct
from flax.core import FrozenDict
from flax.training import train_state
from typing_extensions import Protocol, runtime_checkable

from ciclo.logging import Logs
from ciclo.loops.loop import (
    CallbackOutput,
    FunctionCallbackOutputs,
    LoopCallbackBase,
    LoopState,
    to_standard_outputs,
)
from ciclo.strategies import Strategy, get_strategy
from ciclo.types import Batch, Broadcasts, CluMetric, MetricLike, Statics
from ciclo.utils import inject

Loss = jax.Array

S = TypeVar("S", bound="ManagedState")
A = TypeVar("A")
B = TypeVar("B")

ManagedCallbackCallable = Callable[[Batch, S, Broadcasts, Statics], CallbackOutput[S]]


@runtime_checkable
class HasStrategy(Protocol):
    strategy: Strategy


@runtime_checkable
class HasBatchStats(Protocol):
    batch_stats: Any


@runtime_checkable
class ManagedCallback(Protocol, Generic[S]):
    def __managed_callback__(
        self, state: S, batch: Batch, broadcast: Broadcasts, statics: Statics
    ) -> CallbackOutput[S]:
        ...


# TODO: remove ManagedCallback protocol
@dataclass(frozen=True)
class ManagedFunctionCallback(ManagedCallback[S]):
    f: Callable[..., FunctionCallbackOutputs[S]]

    def __managed_callback__(
        self, state: S, batch: Batch, broadcast: Broadcasts, statics: Statics
    ) -> CallbackOutput[S]:
        outputs = inject(self.f)(state, batch, broadcast, statics)
        return to_standard_outputs(outputs, state)


def _no_apply_given() -> None:
    raise RuntimeError("Trying to use '.apply_fn' but none was passed to `create`.")


class ManagedState(train_state.TrainState):
    """
    A train state that manages the strategy.
    """

    strategy: "Strategy" = struct.field(pytree_node=False)

    @classmethod
    def create(
        cls: Type[S],
        *,
        params: FrozenDict[str, Any],
        tx: optax.GradientTransformation,
        apply_fn: Optional[Callable[..., Any]] = None,
        strategy: Union[Strategy, str] = "jit",
        **kwargs: Any,
    ) -> S:
        apply_fn = apply_fn or _no_apply_given
        state = super().create(
            apply_fn=apply_fn,
            params=params,
            tx=tx,
            strategy=get_strategy("eager"),
            **kwargs,
        )
        return state.with_strategy(strategy)

    def with_strategy(self: S, strategy: Union[Strategy, str]) -> S:
        new_strategy = get_strategy(strategy) if isinstance(strategy, str) else strategy
        current_strategy = self.strategy
        if new_strategy == current_strategy:
            return self
        state = current_strategy.to_host(self)
        state = new_strategy.from_host(state)
        return state.replace(strategy=new_strategy)


@dataclass
class ManagedStep(LoopCallbackBase[S]):
    strategy_callbacks: Dict[Strategy, ManagedCallbackCallable[S]]
    default_strategy: Strategy
    managed_step_fn: ManagedCallback[S]

    def __call__(self, state: S, *args: Any) -> CallbackOutput[S]:
        if len(args) > 3:
            raise ValueError(f"Expected a maximum of 4 arguments, got {len(args) + 1}")

        batch, broadcasts, statics = args + (None,) * (3 - len(args))

        if isinstance(state, HasStrategy):
            strategy = state.strategy
            assert isinstance(strategy, Strategy)
        else:
            strategy = self.default_strategy

        if strategy not in self.strategy_callbacks:
            self.strategy_callbacks[strategy] = strategy(
                self.get_final_callback(strategy)
            )

        callback = self.strategy_callbacks[strategy]

        batch = strategy.lift_batch(batch)
        logs, state = callback(state, batch, broadcasts, statics)

        for collection in logs.keys():
            if collection == "stateful_metrics":
                logs[collection] = strategy.lower_replicated(logs[collection])
            elif collection in ("losses", "metrics"):
                logs[collection] = strategy.lower_averageable(logs[collection])
            elif collection == "outputs":
                logs[collection] = strategy.lower_tileable(logs[collection])
            else:
                logs[collection] = strategy.lower_sharded(logs[collection])

        return logs, state

    def get_final_callback(self, strategy: Strategy) -> ManagedCallbackCallable[S]:
        def lifted_postprocess(
            state: S, batch: Batch, broadcasts: Broadcasts, statics: Statics
        ) -> CallbackOutput[S]:
            step_fn = self.get_step_callback(strategy)
            logs, state = step_fn(state, batch, broadcasts, statics)

            if "stateful_metrics" in logs:
                stateful_metrics = logs["stateful_metrics"]
                assert isinstance(stateful_metrics, MutableMapping)
                for key, value in list(stateful_metrics.items()):
                    if isinstance(value, (CluMetric, MetricLike)):
                        metric = getattr(state, key)
                        assert isinstance(metric, (CluMetric, MetricLike))
                        value = strategy.handle_metric(value)
                        metric = metric.merge(value)
                        state = state.replace(**{key: metric})
                        metric_value = metric.compute()
                        if isinstance(metric_value, Mapping):
                            del stateful_metrics[key]
                            stateful_metrics.update(metric_value)
                        else:
                            stateful_metrics[key] = metric_value
            return logs, state

        return lifted_postprocess

    def get_step_callback(self, strategy: Strategy) -> ManagedCallbackCallable[S]:
        return self.managed_step_fn.__managed_callback__

    def __loop_callback__(self, loop_state: LoopState[S]) -> CallbackOutput[S]:
        logs, state = self(loop_state.state, loop_state.batch, loop_state.elapsed, None)
        return logs, state


@dataclass
class ManagedTrainStep(ManagedStep[S]):
    def get_step_callback(self, strategy: Strategy) -> ManagedCallbackCallable[S]:
        def train_step_callback(
            state: S, batch: Batch, broadcasts: Broadcasts, statics: Statics
        ) -> CallbackOutput[S]:
            def loss_fn(params):
                _state = state.replace(params=params)
                logs, _state = self.managed_step_fn.__managed_callback__(
                    _state, batch, broadcasts, statics
                )
                if "losses" not in logs:
                    raise ValueError(
                        f"callback must return dictorionary with a 'losses' key, but got {logs.keys()}"
                    )
                if len(logs["losses"]) == 0:
                    raise ValueError(
                        "'losses' collection is empty, you must provide at least one entry "
                        "in the 'losses' collection"
                    )

                loss = 0.0
                for k, v in logs["losses"].items():
                    if v.shape != ():
                        raise ValueError(
                            f"Loss {k} should be a scalar, but has shape {v.shape}"
                        )
                    loss += v

                return loss, (logs, _state)

            logs: Logs
            (_, (logs, state)), grads = jax.value_and_grad(loss_fn, has_aux=True)(
                state.params
            )
            grads = strategy.handle_grads(grads)
            state = state.apply_gradients(grads=grads)

            if isinstance(state, HasBatchStats):
                batch_stats = strategy.handle_batch_stats(state.batch_stats)
                state = state.replace(batch_stats=batch_stats)

            return logs, state

        return train_step_callback


def train_step(
    step_fn: Callable[..., Any],
    strategy: Union[Strategy, str] = "jit",
) -> ManagedTrainStep[S]:
    strategy = get_strategy(strategy) if isinstance(strategy, str) else strategy
    return ManagedTrainStep(
        strategy_callbacks={},
        default_strategy=strategy,
        managed_step_fn=ManagedFunctionCallback(step_fn),
    )


def step(
    step_fn: Callable[..., Any],
    strategy: Union[Strategy, str] = "jit",
) -> ManagedStep[S]:
    strategy = get_strategy(strategy) if isinstance(strategy, str) else strategy
    return ManagedStep(
        strategy_callbacks={},
        default_strategy=strategy,
        managed_step_fn=ManagedFunctionCallback(step_fn),
    )



import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import flax.linen as nn
import jax
import jax.numpy as jnp
import jax_metrics as jm
import optax
from flax import struct, traverse_util
from flax.core import FrozenDict
from flax.core.scope import CollectionFilter, DenyList, FrozenVariableDict

import ciclo
from ciclo import managed
from ciclo.strategies import Strategy
from ciclo.timetracking import Elapsed
from ciclo.types import CluMetric, MetricLike

Pytree = Any
Loss = Callable[..., jax.Array]
Index = Union[int, str]
KeyPath = Tuple[Index, ...]
M = TypeVar("M", bound=nn.Module)


def static_field(**kwargs):
    return struct.field(pytree_node=False, **kwargs)


def node_field(**kwargs):
    return struct.field(pytree_node=True, **kwargs)


class Metrics(struct.PyTreeNode, MetricLike):
    metrics: Dict[str, Union[CluMetric, MetricLike]]

    @classmethod
    def create(cls, metrics: Dict[str, Any]):
        # validate metrics
        for name, metric in metrics.items():
            if not isinstance(metric, (CluMetric, MetricLike)):
                raise ValueError(f"Invalid metric type for {name}: {type(metric)}")

        return cls(
            metrics=metrics,
        )

    def reset(self):
        metrics = {}
        for name, metric in self.metrics.items():
            if isinstance(metric, MetricLike):
                metrics[name] = metric.reset()
            else:
                metrics[name] = metric.empty()

        return self.replace(metrics=metrics)

    def update(self, **kwargs):
        metrics = {}
        for name, metric in self.metrics.items():
            if isinstance(metric, MetricLike):
                metrics[name] = metric.update(**kwargs)
            else:
                metrics[name] = metric.merge(metric.from_model_output(**kwargs))

        return self.replace(metrics=metrics)

    def batch_updates(self, **kwargs) -> "Metrics":
        metrics = {}
        for name, metric in self.metrics.items():
            if isinstance(metric, MetricLike):
                metrics[name] = metric.batch_updates(**kwargs)
            else:
                metrics[name] = metric.from_model_output(**kwargs)

        return self.replace(metrics=metrics)

    def merge(self, other: "Metrics") -> "Metrics":
        metrics = {}
        for name, metric in self.metrics.items():
            other_metric = other.metrics[name]
            if type(metric) != type(other_metric):
                raise ValueError(
                    f"Metric {name} has different types: {type(metric)} and "
                    f"{type(other_metric)}"
                )

            if isinstance(metric, MetricLike):
                assert isinstance(other_metric, MetricLike)
                metrics[name] = metric.merge(other_metric)
            else:
                assert isinstance(other_metric, CluMetric)
                metrics[name] = metric.merge(other_metric)

        return self.replace(metrics=metrics)

    def compute(self):
        logs = {}
        for name, metric in self.metrics.items():
            logs[name] = metric.compute()

        return logs

    def reduce(self):
        metrics = {}
        for name, metric in self.metrics.items():
            metrics[name] = metric.reduce()

        return self.replace(metrics=metrics)


class FlaxState(Generic[M], managed.ManagedState):
    """State for Flax models."""

    key: jax.random.KeyArray
    batch_stats: Optional[FrozenVariableDict]
    variables: FrozenVariableDict
    stateful_metrics: Metrics
    stateless_metrics: Dict[str, Callable[..., jax.Array]] = static_field()
    losses: Dict[str, Loss] = static_field()
    module_fn: Callable[[], M] = static_field()

    mutable_train: CollectionFilter = static_field()
    mutable_eval: CollectionFilter = static_field()
    rngs_train: Sequence[str] = static_field()
    rngs_eval: Sequence[str] = static_field()
    method_train: Union[str, Callable[..., Any]] = static_field()
    method_eval: Union[str, Callable[..., Any]] = static_field()
    logs_full_path: bool = static_field()
    keep_collections: Tuple[str, ...] = static_field()

    @property
    def module(self) -> M:
        return self.module_fn()

    def apply(
        self,
        key: jax.random.KeyArray,
        inputs: Any,
        training: bool,
    ):
        method = self.method_train if training else self.method_eval
        method = _unbounded_method(self.module, method)
        mutable = self.mutable_train if training else self.mutable_eval
        rng_names = self.rngs_train if training else self.rngs_eval
        arg_names = _function_argument_names(method)
        args, kwargs = _split_args_kwargs(inputs)

        if (
            arg_names is not None
            and "training" in arg_names
            and "training" not in kwargs
        ):
            kwargs["training"] = training

        rngs = _split_into_collection(key, rng_names)

        variables = self.variables
        if self.params is not None:
            variables = variables.copy({"params": self.params})
        if self.batch_stats is not None:
            variables = variables.copy({"batch_stats": self.batch_stats})

        apply_output = self.module.apply(
            variables,
            *args,
            rngs=rngs,
            mutable=mutable,
            method=method,
            **kwargs,
        )

        if mutable is False:
            outputs = apply_output
        else:
            outputs, variable_updates = apply_output
            variables = variables.copy(variable_updates)

        return outputs, self.replace(
            variables=variables,
            params=variables.get("params", None),
            batch_stats=variables.get("batch_stats", None),
        )

    @managed.train_step
    def train_step(self, batch, elapsed: Elapsed):
        logs, state = self._step(batch, elapsed, training=True)
        return logs, state

    @managed.step
    def test_step(self, batch, elapsed: Elapsed):
        logs, state = self._step(batch, elapsed, training=False)
        return logs, state

    @managed.step
    def predict_step(self, batch, elapsed: Elapsed):
        inputs, labels, sample_weight = unpack_x_y_sample_weight(batch)
        key = jax.random.fold_in(self.key, elapsed.steps)
        preds, _ = self.apply(key, inputs, training=False)

        if isinstance(preds, jax.Array):
            preds = {"preds": preds}
        else:
            preds = traverse_util.flatten_dict(preds, sep="/")

        logs = ciclo.logs()
        for name, value in preds.items():
            logs.add_output(name, value)

        return logs, None

    @managed.step
    def reset_step(self):
        stateful_metrics = self.stateful_metrics.reset()
        return self.replace(stateful_metrics=stateful_metrics)

    def _step(self, batch, elapsed: Elapsed, training: bool):
        inputs, labels, sample_weight = unpack_x_y_sample_weight(batch)

        key = jax.random.fold_in(self.key, elapsed.steps)
        seq_key, pred_key = jax.random.split(key, 2)
        preds, self = self.apply(pred_key, inputs, training=training)
        variables = self.variables

        if not isinstance(labels, Mapping):
            label_args = dict(target=labels)
        else:
            label_args = labels

        if not isinstance(inputs, Mapping):
            input_args = dict(inputs=inputs)
        else:
            input_args = inputs

        arguments = dict(
            preds=preds,
            module=self,
            params=self.params,
            batch_stats=self.batch_stats,
            key_seq=KeySeq(seq_key),
            sample_weight=sample_weight,
            **label_args,
            **input_args,
        )

        logs = ciclo.logs()

        for name, loss in self.losses.items():
            logs.add_loss(name, loss(**arguments))

        if "losses" in variables:
            aux_losses = flatten_names_unique(
                variables["losses"], only_last=not self.logs_full_path
            )
            for name, loss in aux_losses.items():
                loss = jnp.asarray(loss)
                logs.add_loss(name, loss)

        if "losses" not in logs:
            raise ValueError("No losses were added to the logs.")

        # Add losses to metric arguments
        for name, loss in logs.losses.items():
            if name not in arguments:
                arguments[name] = loss

        for name, metric in self.stateless_metrics.items():
            if callable(metric):
                logs.add_metric(name, metric(**arguments))
            else:
                raise TypeError(
                    f"Unexpected metric type {type(metric)} for metric {name}."
                )

        logs.add_stateful_metric(
            "stateful_metrics", self.stateful_metrics.batch_updates(**arguments)
        )

        if "metrics" in variables:
            aux_metrics = flatten_names_unique(
                variables["metrics"], only_last=not self.logs_full_path
            )
            for name, metric in aux_metrics.items():
                metric = jnp.asarray(metric)
                logs.add_metric(name, metric)

        # Filter variables
        variables = FrozenDict(
            {
                collection: variables[collection]
                for collection in variables
                if collection in self.keep_collections
            }
        )

        return logs, self.replace(variables=variables)

    # ---------------------------------------------------------------------------------
    # HighLevel API helpers
    # ---------------------------------------------------------------------------------


def init_module(
    module: nn.Module,
    inputs: Any,
    key: jax.random.KeyArray,
    mutable_init: CollectionFilter,
    rngs_init: Sequence[str],
    method_init: Union[str, Callable[..., Any]],
    init_training_value: bool,
):
    method = _unbounded_method(module, method_init)
    arg_names = _function_argument_names(method)
    args, kwargs = _split_args_kwargs(inputs)

    if arg_names is not None and "training" in arg_names and "training" not in kwargs:
        kwargs["training"] = init_training_value

    rngs = _split_into_collection(key, rngs_init)

    variables = module.init(
        rngs,
        *args,
        mutable=mutable_init,
        method=method,
        **kwargs,
    )
    return variables


def create_flax_state(
    module: M,
    inputs: Any,
    tx: optax.GradientTransformation,
    key: Optional[jax.random.KeyArray] = None,
    losses: Optional[Dict[str, Loss]] = None,
    metrics: Optional[Dict[str, Any]] = None,
    mutable_init: CollectionFilter = True,
    mutable_train: CollectionFilter = DenyList(["params"]),
    mutable_eval: CollectionFilter = DenyList(["params", "batch_stats"]),
    rngs_init: Sequence[str] = ("params", "dropout", "default"),
    rngs_train: Sequence[str] = ("dropout", "default"),
    rngs_eval: Sequence[str] = ("default",),
    method_init: Union[str, Callable[..., Any]] = "__call__",
    method_train: Union[str, Callable[..., Any]] = "__call__",
    method_eval: Union[str, Callable[..., Any]] = "__call__",
    init_training_value: bool = True,
    logs_full_path: bool = False,
    keep_collections: Sequence[str] = (),
    strategy: Union[Strategy, str] = "jit",
) -> FlaxState[M]:
    if key is None:
        key = jax.random.PRNGKey(0)
    if losses is None:
        losses = {}
    if metrics is None:
        metrics = {}

    keep_collections = tuple(keep_collections)

    variables = init_module(
        module=module,
        inputs=inputs,
        key=key,
        mutable_init=mutable_init,
        rngs_init=rngs_init,
        method_init=method_init,
        init_training_value=init_training_value,
    )
    if "params" in variables:
        variables, params = variables.pop("params")
    else:
        params = None

    if "batch_stats" in variables:
        variables, batch_stats = variables.pop("batch_stats")
    else:
        batch_stats = None

    # Create metrics
    stateful_metrics = {}
    stateless_metrics = {}

    # Split metrics into stateful and stateless metrics
    for name, metric in metrics.items():
        if isinstance(metric, (CluMetric, MetricLike)):
            stateful_metrics[name] = metric
        elif issubclass(metric, CluMetric):
            stateful_metrics[name] = metric.empty()
        elif callable(metric):
            stateless_metrics[name] = metric
        else:
            raise ValueError(f"Invalid metric type for {name}: {type(metric)}")

    # Add stateful metrics for auxiliary losses
    if "losses" in variables:
        aux_losses = flatten_names_unique(
            variables["losses"], only_last=not logs_full_path
        )
        for name in aux_losses.keys():
            avg_name = f"avg_{name}"
            if avg_name not in stateful_metrics:
                stateful_metrics[avg_name] = jm.metrics.Mean().from_argument(name)

    # Add stateful metrics for losses
    for name in losses.keys():
        avg_name = f"avg_{name}"
        if avg_name not in stateful_metrics:
            stateful_metrics[avg_name] = jm.metrics.Mean().from_argument(name)

    # Filter variables
    variables = FrozenDict(
        {
            collection: variables[collection]
            for collection in variables
            if collection in keep_collections
        }
    )

    state = FlaxState.create(
        # TrainState
        apply_fn=module.apply,
        params=params,
        tx=tx,
        # ManagedState
        strategy=strategy,
        # FlaxState
        key=key,
        batch_stats=batch_stats,
        variables=variables,
        losses=losses,
        stateful_metrics=Metrics.create(stateful_metrics),
        stateless_metrics=stateless_metrics,
        module_fn=lambda: module,
        mutable_train=mutable_train,
        mutable_eval=mutable_eval,
        rngs_train=rngs_train,
        rngs_eval=rngs_eval,
        method_train=method_train,
        method_eval=method_eval,
        logs_full_path=logs_full_path,
        keep_collections=keep_collections,
    )
    return state


def _function_argument_names(f: Callable) -> Optional[List[str]]:
    """
    Returns:
        A list of keyword argument names or None if variable keyword arguments (`**kwargs`) are present.
    """
    kwarg_names = []

    for k, v in inspect.signature(f).parameters.items():
        if v.kind == inspect.Parameter.VAR_KEYWORD:
            return None

        kwarg_names.append(k)

    return kwarg_names


def _split_args_kwargs(
    value: Any,
) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    if isinstance(value, tuple):
        return value, {}
    elif isinstance(value, list):
        return tuple(value), {}
    elif isinstance(value, Mapping):
        return (), dict(value)
    else:
        return (value,), {}


def _split_into_collection(
    key: jnp.ndarray,
    collection: Sequence[str],
) -> Dict[str, jnp.ndarray]:
    """
    Split the key into the specified rngs.
    """

    keys = jax.random.split(key, len(collection))

    keys_collection = {col: keys[i] for i, col in enumerate(collection)}

    return keys_collection


def _unbounded_method(
    module: nn.Module,
    method: Union[str, Callable[..., Any]],
) -> Callable[..., Any]:
    if isinstance(method, str):
        return getattr(type(module), method)
    return method


def unpack_x_y_sample_weight(data):
    """Unpacks user-provided data tuple."""
    if not isinstance(data, tuple):
        return (data, None, None)
    elif len(data) == 1:
        return (data[0], None, None)
    elif len(data) == 2:
        return (data[0], data[1], None)
    elif len(data) == 3:
        return (data[0], data[1], data[2])

    raise ValueError("Data not understood.")


def flatten_names_unique(inputs: Any, only_last: bool = False) -> Dict[str, Any]:
    names: Set[str] = set()

    if only_last:
        return {
            get_unique_name(names, str(path[-1])): value
            for path, value in _flatten_names((), inputs)
        }
    else:
        return {
            get_unique_name(names, "/".join(map(str, path))): value
            for path, value in _flatten_names((), inputs)
        }


def get_unique_name(
    names: Set[str],
    name: str,
):
    if name in names:
        i = 1
        while f"{name}_{i}" in names:
            i += 1

        name = f"{name}_{i}"

    names.add(name)
    return name


def _flatten_names(path: KeyPath, inputs: Any) -> Iterable[Tuple[KeyPath, Any]]:
    if isinstance(inputs, (Tuple, List)):
        for i, value in enumerate(inputs):
            yield from _flatten_names(path, value)
    elif isinstance(inputs, Mapping):
        for name, value in inputs.items():
            yield from _flatten_names(path + (name,), value)
    else:
        yield (path, inputs)


class KeySeq:
    """KeySeq is simple module that can produce a sequence of PRNGKeys.

    Example:
    ```python
    class Dropout(Module):
        rng: KeySeq()

        def __init__(self, rate: float):
            self.next_key = KeySeq()
            ...

        def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
            key = self.next_key()
            mask = jax.random.bernoulli(key, 1.0 - self.rate)
            ...
    ```
    """

    key: jax.random.KeyArray
    index: int

    def __init__(
        self,
        key: Union[jax.random.KeyArray, int],
    ):
        """
        Arguments:
            key: An optional PRNGKey to initialize the KeySeq with.
        """

        self.key = jax.random.PRNGKey(key) if isinstance(key, int) else key
        self.index = 0

    def next(self) -> jax.random.KeyArray:
        """
        Return a new PRNGKey and updates the internal rng state.

        Returns:
            A PRNGKey.
        """

        key = jax.random.fold_in(self.key, self.index)
        self.index += 1
        return key

    __next__ = next
